use std::{
    collections::{HashMap, HashSet},
    error::Error,
    sync::mpsc,
};

use itertools::Itertools;
use rayon::prelude::*;

pub fn process_part1(input: &str) -> u32 {
    simulate_all_2(input, 100).values().copied().sum()
    //simulate_all(input)
    //    .iter()
    //    .map(
    //        |(time_saved, count)| {
    //            if *time_saved >= 100 {
    //                *count
    //            } else {
    //                0
    //            }
    //        },
    //    )
    //    .sum()
}

fn simulate_all(input: &str) -> HashMap<u32, u32> {
    let mut cheats = Vec::new();
    let mut start = (0, 0);
    let grid = input
        .lines()
        .enumerate()
        .map(|(yidx, row)| {
            row.chars()
                .enumerate()
                .map(|(xidx, chara)| {
                    let tile = GridTile::try_from(chara).unwrap();
                    if tile == GridTile::Start {
                        start = (xidx, yidx);
                    }
                    tile
                })
                .collect_vec()
        })
        .collect_vec();
    grid[1..grid.len() - 1]
        .iter()
        .enumerate()
        .for_each(|(yidx, row)| {
            row[1..row.len() - 1]
                .iter()
                .enumerate()
                .for_each(|(xidx, tile)| {
                    if *tile == GridTile::Wall && cheatable(&grid, (xidx + 1, yidx + 1)) {
                        cheats.push((xidx + 1, yidx + 1));
                    }
                });
        });
    let mut saved = HashMap::new();

    let no_cheat = simulate(grid.clone(), start, None).unwrap();
    let (tx, rx) = mpsc::channel();
    cheats.into_par_iter().for_each(|cheat| {
        let cheat = simulate(grid.clone(), start, Some(cheat)).unwrap();
        let time_saved = no_cheat - cheat;
        let _ = tx.send(time_saved);
    });
    drop(tx);
    while let Ok(time_saved) = rx.recv() {
        saved
            .entry(time_saved)
            .and_modify(|count| *count += 1)
            .or_insert(1);
    }
    saved
}

fn cheatable(grid: &[Vec<GridTile>], wall_position: (usize, usize)) -> bool {
    let above = grid
        .get(wall_position.1 - 1)
        .map(|above_row| above_row[wall_position.0]);
    let below = grid
        .get(wall_position.1 + 1)
        .map(|below_row| below_row[wall_position.0]);
    let left = grid[wall_position.1].get(wall_position.0 - 1).copied();
    let right = grid[wall_position.1].get(wall_position.0 + 1).copied();
    [above, below, left, right]
        .iter()
        .map(|tile| {
            if let Some(tile) = tile {
                match tile {
                    GridTile::Wall => 0,
                    GridTile::Path | GridTile::Start | GridTile::End => 1,
                }
            } else {
                0
            }
        })
        .sum::<u32>()
        >= 2
}

fn simulate(
    mut grid: Vec<Vec<GridTile>>,
    start: (usize, usize),
    cheat: Option<(usize, usize)>,
) -> Option<u32> {
    if let Some(cheat) = cheat {
        grid[cheat.1][cheat.0] = GridTile::Path;
    }
    let mut visited = HashSet::new();
    visited.insert(start);
    let mut next_paths = vec![MazeRunner {
        coords: start,
        visited: visited.clone(),
        ..Default::default()
    }];
    let mut arrived: Vec<MazeRunner> = Vec::new();
    while !next_paths.is_empty() {
        next_paths = next_paths
            .iter()
            .map(|maze_runner| {
                let mut paths = Vec::new();
                if let Some(path) = maze_runner.get_next(&grid, Direction::Up) {
                    paths.push(path);
                }
                if let Some(path) = maze_runner.get_next(&grid, Direction::Down) {
                    paths.push(path);
                }
                if let Some(path) = maze_runner.get_next(&grid, Direction::Left) {
                    paths.push(path);
                }
                if let Some(path) = maze_runner.get_next(&grid, Direction::Right) {
                    paths.push(path);
                }
                paths
            })
            .collect_vec()
            .concat();
        for (idx, maze_runner) in next_paths.clone().iter().enumerate().rev() {
            if maze_runner.visited.contains(&maze_runner.coords)
                || visited.contains(&maze_runner.coords)
            {
                next_paths.remove(idx);
                continue;
            }
            visited.insert(maze_runner.coords);
            next_paths[idx].visited.insert(maze_runner.coords);
            if maze_runner.state == State::Arrived {
                let arrived_reindeer = next_paths.remove(idx);
                arrived.push(arrived_reindeer);
            }
            //log_maze(&grid, maze_runner);
        }
    }
    arrived.sort_by(|a_runner, b_runner| a_runner.visited.len().cmp(&b_runner.visited.len()));
    arrived
        .first()
        .map(|arrived| arrived.visited.len() as u32 - 1)
}

pub fn process_part2(input: &str) -> u32 {
    simulate_all_2(input, 100).values().copied().sum()
}

fn simulate_all_2(input: &str, time_to_saved: usize) -> HashMap<u32, u32> {
    let mut cheats = Vec::new();
    let mut start = (0, 0);
    let grid = input
        .lines()
        .enumerate()
        .map(|(yidx, row)| {
            row.chars()
                .enumerate()
                .map(|(xidx, chara)| {
                    let tile = GridTile::try_from(chara).unwrap();
                    if tile == GridTile::Start {
                        start = (xidx, yidx);
                    }
                    tile
                })
                .collect_vec()
        })
        .collect_vec();
    grid[1..grid.len() - 1]
        .iter()
        .enumerate()
        .for_each(|(yidx, row)| {
            row[1..row.len() - 1]
                .iter()
                .enumerate()
                .for_each(|(xidx, tile)| {
                    if *tile == GridTile::Wall && cheatable(&grid, (xidx + 1, yidx + 1)) {
                        cheats.push((xidx + 1, yidx + 1));
                    }
                });
        });
    let no_cheat = simulate_2(&grid, start);
    let mut saved = HashMap::new();
    for (tile_idx, tile) in no_cheat[..no_cheat.len() - time_to_saved]
        .iter()
        .enumerate()
    {
        for (cheat_idx, cheat) in no_cheat[tile_idx + time_to_saved..].iter().enumerate() {
            let manhattan = tile.0.abs_diff(cheat.0) + tile.1.abs_diff(cheat.1);
            if manhattan <= 2 {
                let time_saved = (time_to_saved + cheat_idx).saturating_sub(2);
                if time_saved != 0 {
                    saved
                        .entry(time_saved as u32)
                        .and_modify(|count| *count += 1)
                        .or_insert(1);
                }
            }
        }
    }
    saved
}

fn simulate_2(grid: &[Vec<GridTile>], start: (usize, usize)) -> Vec<(usize, usize)> {
    let mut visited = HashSet::new();
    visited.insert(start);
    let mut next_paths = vec![MazeRunner {
        coords: start,
        visited: visited.clone(),
        ..Default::default()
    }];
    let mut visited = Vec::new();
    visited.push(start);
    let mut arrived: Vec<MazeRunner> = Vec::new();
    while !next_paths.is_empty() {
        next_paths = next_paths
            .iter()
            .map(|maze_runner| {
                let mut paths = Vec::new();
                if let Some(path) = maze_runner.get_next(grid, Direction::Up) {
                    paths.push(path);
                }
                if let Some(path) = maze_runner.get_next(grid, Direction::Down) {
                    paths.push(path);
                }
                if let Some(path) = maze_runner.get_next(grid, Direction::Left) {
                    paths.push(path);
                }
                if let Some(path) = maze_runner.get_next(grid, Direction::Right) {
                    paths.push(path);
                }
                paths
            })
            .collect_vec()
            .concat();
        for (idx, maze_runner) in next_paths.clone().iter().enumerate().rev() {
            if maze_runner.visited.contains(&maze_runner.coords)
                || visited.contains(&maze_runner.coords)
            {
                next_paths.remove(idx);
                continue;
            }
            visited.push(maze_runner.coords);
            next_paths[idx].visited.insert(maze_runner.coords);
            if maze_runner.state == State::Arrived {
                let arrived_reindeer = next_paths.remove(idx);
                arrived.push(arrived_reindeer);
            }
            //log_maze(&grid, maze_runner);
        }
    }
    visited
}

#[derive(Debug, Clone, PartialEq, Eq)]
enum Direction {
    Up,
    Down,
    Left,
    Right,
}

#[derive(Debug, Default, Clone, PartialEq, Eq)]
enum State {
    #[default]
    Going,
    Arrived,
}

#[derive(Debug, Default, Clone, PartialEq, Eq)]
struct MazeRunner {
    coords: (usize, usize),
    visited: HashSet<(usize, usize)>,
    state: State,
}

impl MazeRunner {
    fn get_next(&self, grid: &[Vec<GridTile>], direction: Direction) -> Option<MazeRunner> {
        let tile = match direction {
            Direction::Up => grid[self.coords.1 - 1][self.coords.0],
            Direction::Down => grid[self.coords.1 + 1][self.coords.0],
            Direction::Right => grid[self.coords.1][self.coords.0 + 1],
            Direction::Left => grid[self.coords.1][self.coords.0 - 1],
        };
        let coords = match direction {
            Direction::Up => (self.coords.0, self.coords.1 - 1),
            Direction::Down => (self.coords.0, self.coords.1 + 1),
            Direction::Left => (self.coords.0 - 1, self.coords.1),
            Direction::Right => (self.coords.0 + 1, self.coords.1),
        };
        if tile == GridTile::Wall {
            None
        } else if tile == GridTile::End {
            Some(MazeRunner {
                state: State::Arrived,
                coords,
                ..self.clone()
            })
        } else {
            Some(MazeRunner {
                coords,
                ..self.clone()
            })
        }
    }
}

#[derive(Debug, Default, Clone, Copy, Hash, PartialEq, Eq)]
enum GridTile {
    Wall,
    #[default]
    Path,
    Start,
    End,
}

impl TryFrom<char> for GridTile {
    type Error = Box<dyn Error>;

    fn try_from(value: char) -> std::result::Result<GridTile, Box<dyn Error>> {
        match value {
            '#' => Ok(Self::Wall),
            '.' => Ok(Self::Path),
            'S' => Ok(Self::Start),
            'E' => Ok(Self::End),
            _ => Err(Box::from(format!("{value} is not a valid tile"))),
        }
    }
}

fn log_maze(grid: &[Vec<GridTile>], maze_runner: &MazeRunner) {
    for (yidx, row) in grid.iter().enumerate() {
        for (xidx, tile) in row.iter().enumerate() {
            let contains = {
                maze_runner.visited.contains(&(xidx, yidx))
                    || maze_runner.visited.contains(&(xidx, yidx))
                    || maze_runner.visited.contains(&(xidx, yidx))
                    || maze_runner.visited.contains(&(xidx, yidx))
            };
            if contains && !(*tile == GridTile::Start || *tile == GridTile::End) {
                print!("O");
            } else if *tile == GridTile::Wall {
                print!("#");
            } else if *tile == GridTile::Path {
                print!(".");
            } else if *tile == GridTile::Start {
                print!("S");
            } else if *tile == GridTile::End {
                print!("E");
            }
        }
        println!();
    }
}

#[cfg(test)]
mod tests {
    use std::time::Instant;

    use utils::time::get_elapsed_string;

    use super::*;

    const INPUT: &str = "###############
#...#...#.....#
#.#.#.#.#.###.#
#S#...#.#.#...#
#######.#.#.###
#######.#.#...#
#######.#.###.#
###..E#...#...#
###.#######.###
#...###...#...#
#.#####.#.###.#
#.#...#.#.#...#
#.#.#.#.#.#.###
#...#...#...###
###############";

    #[test]
    fn part1() {
        let now = Instant::now();
        println!("Test 1:");
        let result = simulate_all(INPUT);
        result
            .iter()
            .sorted_by(|a, b| a.0.cmp(b.0))
            .for_each(|(saved, count)| {
                println!("There are {count} cheats that saved {saved} picoseconds");
            });
        println!("Ran in {}", get_elapsed_string(now.elapsed()));
    }

    #[test]
    fn part1_fast() {
        let now = Instant::now();
        println!("Test 2:");
        let result = simulate_all_2(INPUT, 0);
        result
            .iter()
            .sorted_by(|a, b| a.0.cmp(b.0))
            .for_each(|(saved, count)| {
                println!("There are {count} cheats that saved {saved} picoseconds");
            });
        println!("Ran in {}", get_elapsed_string(now.elapsed()));
    }

    #[test]
    fn part2() {
        let now = Instant::now();
        println!("Test 2:");
        let result = simulate_all_2(INPUT, 2);
        result
            .iter()
            .sorted_by(|a, b| a.0.cmp(b.0))
            .for_each(|(saved, count)| {
                println!("There are {count} cheats that saved {saved} picoseconds");
            });
        println!("Ran in {}", get_elapsed_string(now.elapsed()));
    }
}
